module NDerive
	contains
		function D(f, x, eps, method, maxIter)
			implicit none
			real(kind = 8) :: D
			real(kind = 8) , intent(in) :: x
			real(kind = 8) , intent(in) , optional :: eps
			real(kind = 4) , intent(in) , optional :: maxIter
			optional :: method
				interface
					function f(x)
						implicit none
						real(kind = 8) :: f
						real(kind = 8) , intent(in) :: x
					end function f
					function method(x, h, f)
						implicit none
						real(kind = 8) :: method
						real(kind = 8) , intent(in) :: x, h
							interface
								function f(x)
									implicit none
									real(kind = 8) :: f
									real(kind = 8) , intent(in) :: x
								end function f
							end interface
					end function method
				end interface
				
				real(kind = 8) :: tol, step
				real(kind = 8) :: derPrev, derCurr, derNext
				integer(kind = 4) :: max, i
				logical:: ok
					if (present(eps)) then
						tol = eps
						else
						tol = 1.0d-7
					end if
					if (present(maxIter)) then
						max = maxIter
						else
						max = 100
					end if
						step = 1.0d-1*x
					if (present(method)) then
						derPrev = method(x, step, f)
						step = 1.0e-1_8*step
						derCurr = method(x, step, f)
						i = 0
					do
						step = 1.0e-1_8*step
						derNext = method(x, step , f)
						ok = (abs(derNext - derCurr)) >=abs(derCurr - derPrev).OR.(abs(derCurr - derPrev) < tol)
					if(i > max) stop 'ERROR!!! Too Many Steps in D calculation'
						i = i + 1
						derPrev = derCurr
						derCurr = derNext
					if(ok) exit
					end do
					else
							derPrev = derDet(x, step)
							step = 1.0e-1_8*step
							derCurr = derDet(x, step)
						i = 0
						do 
							step = 1.0e-1_8*step
							derNext = derDet(x, step)
							ok = (abs(derNext - derCurr)) >=abs(derCurr - derPrev).OR.(abs(derCurr - derPrev) < tol)
						if(i > max) stop 'ERROR!!! Too Many Steps in D calculation'
							i = i + 1
							derPrev = derCurr
							derCurr = derNext
							if(ok) exit
						end do
				    end if 
				D = derCurr
			contains
				function derDet(x, h)
						real(kind = 8) :: derDet
						real(kind = 8) , intent(in) ::  x, h
						DerDet = (f(x + h) - f(x)) / h
				end function derDet
		end function D	
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	function der3ptBackward(x, h, f)
	implicit none
		real(kind = 8) :: der3ptBackward
		real(kind = 8) , intent(in) :: x, h
			interface
				function f(x)
					implicit none
					real(kind = 8) :: f
					real(kind = 8) , intent(in) :: x
				end function f
			end interface		
		der3ptBackward = (3.0d0*f(x) + f(-2.0d0*h + x) - 4.0d0*f(-h + x))/(2.0d0*h)
	end function der3ptBackward
!!!
	function der3ptForward(x, h, f)
	implicit none
		real(kind = 8) :: der3ptForward
		real(kind = 8) , intent(in) :: x, h
			interface
				function f(x)
					implicit none
					real(kind = 8) :: f
					real(kind = 8) , intent(in) :: x
				end function f
			end interface
		der3ptForward = -(3.0d0*f(x) - 4.0d0*f(h + x) + f(2.0d0*h + x))/(2.0d0*h)
	end function der3ptForward
!!!
	function der3ptCenter(x, h, f)
	implicit none
		real(kind = 8) :: der3ptCenter
		real(kind = 8) , intent(in) :: x, h
			interface
				function f(x)
					implicit none
					real(kind = 8) :: f
					real(kind = 8) , intent(in) :: x
				end function f
			end interface
		der3ptCenter = (-f(-h + x) + f(h + x))/(2.0d0*h)
	end function der3ptCenter
!!!
		function der5ptBackward(x, h, f)
	implicit none
		real(kind = 8) :: der5ptBackward
		real(kind = 8) , intent(in) :: x, h
			interface
				function f(x)
					implicit none
					real(kind = 8) :: f
					real(kind = 8) , intent(in) :: x
				end function f
			end interface		
		der5ptBackward = (25.0d0*f(x) + 3.0d0*f(-4.0d0*h + x) - 16.0d0*f(-3.0d0*h + x) + 36.0d0*f(-2.0d0*h + x) - 48.0d0*f(-h + x))/(12.*h)
	end function der5ptBackward
!!!
	function der5ptForward(x, h, f)
	implicit none
		real(kind = 8) :: der5ptForward
		real(kind = 8) , intent(in) :: x, h
			interface
				function f(x)
					implicit none
					real(kind = 8) :: f
					real(kind = 8) , intent(in) :: x
				end function f
			end interface
		der5ptForward = -(25.0d0*f(x) - 48.0d0*f(h + x) + 36.0d0*f(2.0d0*h + x) - 16.0d0*f(3.0d0*h + x) + 3.0d0*f(4.0d0*h + x))/(12.0d0*h)
	end function der5ptForward
!!!
	function der5ptCentr(x, h, f)
	implicit none
		real(kind = 8) :: der5ptCentr
		real(kind = 8) , intent(in) :: x, h
			interface
				function f(x)
					implicit none
					real(kind = 8) :: f
					real(kind = 8) , intent(in) :: x
				end function f
			end interface
		der5ptCentr = (f(-2.0d0*h + x) - 8.0d0*f(-h + x) + 8.0d0*f(h + x) - f(2.0d0*h + x))/(12.0d0*h)
	end function der5ptCentr
end module NDerive
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
module Interf
	interface
		function fun(x)
			implicit none
			real(kind = 8) :: fun
			real(kind = 8) , intent(in) :: x
		 end function fun
	end interface
end module Interf
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
program TaskDerivMain
	use NDerive
	use Interf
	implicit none
		real(kind = 8) :: arg, resNumer, resAnal
			arg = 1.0e0_8
			resNumer = D(fun, arg)
			resAnal = 2.0d0*arg - exp(arg)
		print*
		print*, 'Numerical Value ', sngl(resNumer)
		print*, 'Analitical Value ', sngl(resAnal)
		print*, 'Absolute Error ', sngl(abs(resNumer - resAnal))
		print*, 'Rerative Error ', sngl(abs(resNumer - resAnal) / resAnal)
		print*
		print*, 'Next form for invoke: ', sngl(D(fun, arg, method = der3ptCenter))
end program TaskDerivMain
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
function fun(x)
	implicit none
	real(kind = 8) :: fun
	real(kind = 8) , intent(in) :: x
	fun = x*x - exp(x)

	pause
end  function fun