      module NSolve

	  contains
	    !--------------------------------------------
		function dihotomySolve(f, a, b, eps)
		  external f

		  real(kind = 8), intent(in):: a, b, eps
		  real(kind = 8):: x1, x2, temp, half, curr_f

		  x1 = a
		  x2 = b
		  if ( ((f(x1) > 0).AND.(f(x2) > 0)).OR.((f(x1) < 0).AND.(f(x2) < 0)) ) then
			print *, "No solutions :'("
			return
		  else 
		    !   x1   a, b;  x2 - 
		    if (x1 > x2) then 
			  temp = x1
			  x1 = x2
			  x2 = temp
			endif
			!    ,         
			do while (abs(x1 - x2) > eps)
			  half = (x2+x1)/2
			  print *, x1, "|", x2
			  curr_f = f(half)

			  if ( abs(curr_f) == 0.0 ) then
			    res = half
				return
			  endif

			  if (((curr_f > 0).AND.(f(x1) > 0)).OR.((curr_f < 0).AND.(f(x1) < 0)))  then 
			    x1 = half
			  else 
			    x2 = half
			  endif
			end do 
			dihotomySolve = half
		  endif

		end function dihotomySolve
		!--------------------------------------------
		function newtonSolve(f, x0, eps)
		  use NDeriv
		  external f

		  real(kind = 8), intent(in):: x0, eps
		  real(kind = 8):: x1, x2, h

		  h = 0.01
		  x1 = x0
		  x2 = x1 - f(x1)/der3t(f, x1, h, eps)
		  !   ,      
		  do while (abs(x2 - x1) > eps)
			x1 = x2
			x2 = x1 - f(x1)/der3t(f, x1, h, eps)
		  end do 
		  newtonSolve = x2
		end function newtonSolve
        !--------------------------------------------
		function secantSolve(f, x0, eps)
		  use NDeriv
		  external f

		  real(kind = 8), intent(in):: x0, eps
		  real(kind = 8):: x1, x2, h, l0, deriv
		  h = 0.01
		  deriv = der3t(f, x0, h, eps)
		  l0 = 3/deriv !      |l0|   >   2 / f'(x) ; sign(l0) = sign(f'(x))
		  x1 = x0
		  x2 = x1 - l0 * f(x1)
		  !   ,      
		  do while (abs(x2 - x1) > eps)
			x1 = x2
			x2 = x1 - l0 * f(x1)
		  end do 
		  secantSolve = x2
		end function secantSolve
        !--------------------------------------------
		function quadrSolve(f, x0, eps)
		  external f

		  real(kind = 8), intent(in):: x0, eps
		  real(kind = 8):: x, x1, x2, temp
		  real(kind = 8):: q, a, b, c, sq

		  x2 = -1
		  x1 = 1
		  x  = 0

		  !   ,       
		  do while (abs(f(x)) > eps)
		    q = (x - x1) / (x1 - x2)
			a = q*f(x) - q*(q + 1)*f(x1) + (q**2)*f(x2)
			b = (2*q + 1)*f(x) - ((1 + q)**2)*f(x1) + (q**2)*f(x2)
			c = (1 + q)*f(x)
			temp = x
			sq = (b**2 - 4.0*a*c)
			if (sq >= 0) then
			  sq = sq**(1.0/2)*b/abs(b)
			else 
			  quadrSolve = 0
			  return
			end if
			x = x - (x - x1)*(2.0*c/(b + sq))
			x2 = x1
			x1 = temp
		  end do 
		  quadrSolve = x
		end function quadrSolve


	  end module NSolve
